# sqltail db

import dotmap
import logging
import mysql.connector
import os
import sys
import traceback
import configparser
import pathlib

class DatabaseException(Exception):
    """base class for Database exceptions""" 
    pass
               
class DatabaseConnectionFailed(DatabaseException):
    pass

class DatabaseNotFound(Exception):
    """Raised when the expected database is not present."""
    pass


class Database():

    def __init__(self, host=None, port=None, user=None, password=None, database=None, config_file=None, debug=False, verbose=False, suffix=''):

        self.cxn = None
        self.debug = debug
        self.verbose = verbose 
        self.logger = logging.getLogger(self.__class__.__name__)
        if debug:
            self.logger.setLevel(logging.DEBUG)
        self.cfg=self.init_config(config_file)
        self.host = self.get_parameter(host, 'DB_HOST', 'client', 'host')
        self.port = self.get_parameter(port, 'DB_PORT', 'client', 'port')
        self.user = self.get_parameter(user, 'DB_USER', 'client', 'user')
        self.password = self.get_parameter(password, 'DB_PASSWORD', 'client', 'password')
        self.database = self.get_parameter(database, 'DB_DATABASE', 'client', 'database')
        if suffix:
            self.database += suffix

        # connect without specifying a database
        if not self.connect(database=None):
            raise DatabaseConectionFailed(f"Failed connection to database {self.connection_string}")

        # verify the expected database is present
        with self.cursor(tuple=True) as cursor:
            rows = cursor.query('show databases;')
            available_databases = [r[0] for r in rows]
        self.logger.debug(f"databases found: {repr(available_databases)}")

        if not self.database in available_databases:
            raise DatabaseNotFound(f"Database {self.database} is not present.")

        # specify the database we need
        self.cxn.database = self.database;

        self.logger.debug(f"{self}")

    def init_config(self, filename):
        cfg = configparser.ConfigParser()
        if filename:
            path = [pathlib.Path(filename)]
        else:
            path = [
                pathlib.Path().home() / '.sqltail',
                pathlib.Path().home() / '.my.cnf'
            ]
        cfg.read(path)
        return cfg

    def get_parameter(self, value, env, section, key):
        if not value:
            value = os.environ.get(env)
            if not value:
                value = self.cfg[section][key]
        return value


    def __enter__(self):
        return self

    def __exit__(self, etype, value, tb):
        self.close()

    def __del__(self):
        self.close()

    def close(self):
        if self.cxn:
            self.cxn.close()
            self.cxn = None

    def __str__(self):
        """return string representaton for this instance"""
        return f'{self.__class__.__name__}<{self.connection_string} {self.cxn}>'

    def cursor(self, **kwargs):
        """return a Cursor configured with the database connection"""
        return Cursor(self, **kwargs)

    def connect(self, database=None):
        self.connection_string = f"mysql://{self.user}@{self.host}:{self.port}/{self.database if self.database else ''}"
        self.cxn = mysql.connector.connect(
            host=self.host, port=int(self.port), user=self.user, password=self.password, database=database,
            consume_results=True
        )
        self.cxn.get_warnings = True
        return self.cxn


"""
The Cursor() class wraps the mysql.connector.connection.cursor() class in a
context manager.  It supports execute() and query() functions, with local
exception handling passing diagnostic and debugging data to the logging 
system on errors.

Example: 

    with Cursor(db) as cursor:
        rows = cursor.execute('show databases;')

    for row in rows:
        print(repr(row))
    
"""


class Cursor():
    def __init__(self, db, **kwargs):
        """
        Cursor() Constructor

        :db: the parent Database instance
        :param **kwargs: keyword arguments (see below)
        :return: returns nothing
        
        The following keyword arguments may be passed boolean values:

        commit - call commit() after each statement
        ignore_warnings - query() function will request and log MySQL warnings
        ignore_notes - query() function will ignore 'Note' type warnings
        dictionary - query() will return rows as type dict
        tuple - query() will return rows as type tuple
        buffered - passed to the cursor() constructor to modify its function (see MySQL documentation)
        """
        self.db = db
        self.logger = logging.getLogger(self.__class__.__name__)
        if self.db.debug:
            self.logger.setLevel(logging.DEBUG)
        self.sql = ''
        self.lastrowid = None
        self.commit = kwargs.get('commit', False)
        self.ignore_warnings = kwargs.get('ignore_warnings', False)
        self.ignore_notes = kwargs.get('ignore_notes', True)
        self.buffered = kwargs.get('buffered', False)
        self.dictionary = kwargs.get('dictionary', False)
        self.tuple = kwargs.get('tuple', False)
        self.return_rows = True
        if self.tuple:
            self.dictionary = False
            self.return_rows = False 
        elif self.dictionary:
            self.return_rows = False 
        else:
            self.dictionary = True
            self.return_rows = True
        self.cursor = self.db.cxn.cursor(dictionary=self.dictionary, buffered=self.buffered)
        self.logger.debug(f"{self}")

    def __str__(self):
        return f"{__class__.__name__}{self.cursor}"

    def __enter__(self):
        return self

    def __exit__(self, etype, value, tb):
        if self.commit and self.db.cxn.in_transaction:
            self.db.cxn.commit()
        self.cursor.close()

    def _execute(self, *args, **kwargs):
        sql = repr(args)
        self.with_rows = False
        self.column_names = None
        self.description = None
        self.lastrowid = None
        self.rowcount = 0
        self.statement=None
        try:
            result = self.cursor.execute(*args, **kwargs)
            assert result == None
            self.with_rows = self.cursor.with_rows
            self.column_names = self.cursor.column_names
            self.description = self.cursor.description
            self.lastrowid = self.cursor.lastrowid
            self.rowcount = self.cursor.rowcount
            self.statement = self.cursor.statement
            if self.commit and self.db.cxn.in_transaction:
                self.db.cxn.commit()
        except mysql.connector.Error as e:
            f = traceback.extract_stack()[-3]
            self.logger.error(f"SQL Error {e} caller={f.filename}:{f.lineno} query={sql}")
            if self.db.verbose:
                self._dump_sql_error(*args)
            raise e
        return self.cursor

    def _dump_sql_error(self, sql):
        sys.stderr.write('SQL_ERROR_IN:\n' + sql + '\n')
        sys.stderr.flush()

    def execute(self, *args, **kwargs):
        self.logger.debug(f"{self} {args} {kwargs}")
        ret = self._execute(*args, **kwargs)
        self.logger.debug(f"{self} returning {ret}")
        return self 

    def query(self, *args, **kwargs):
        self.logger.debug(f"{self} {args} {kwargs}")
        rows = self._execute(*args, **kwargs).fetchall()
        if not self.ignore_warnings:
            self.handle_warnings()
        if self.return_rows:
            rows = [Row(row) for row in rows]
        self.logger.debug(f"{self} returning {len(rows)} {'row' if len(rows)==1 else 'rows'}")
        if self.db.verbose:
            for i, row in enumerate(rows):
                self.logger.debug('row[{i}] {row}')
        return rows

    def commit(self):
        if self.db.cxn.in_transaction:
            self.db.cxn.commit()
        else:
            self.logger.warning('COMMIT requested while not in transaction')

    def handle_warnings(self):
        for warning in self.cursor.fetchwarnings() or []:
            if warning[0] == 'Note' and self.ignore_notes:
                pass
            else:
                self.logger.warning(f"SQL Warning: {warning}")


class Row(dict):

    def __getattr__(self, key):
        return super().__getitem__(key)

    def __setitem__(self, key, value):
        raise KeyError('Row is read-only')

    def __setattr__(self, key, value):
        raise KeyError('Row is read-only')



# cli

import arrow
import click
import json
import logging
import time
from pathlib import Path

from sqltail import SQLTail, Database, DatabaseNotFound, DatabaseConnectionFailed, __version__, __license__

@click.command(name='sqltail')
@click.version_option(message=f"sqltail v{__version__} {__license__}")
@click.option('--host', envvar='DB_HOST', type=str)
@click.option('--port', envvar='DB_PORT', type=str)
@click.option('--user', envvar='DB_USER', type=str)
@click.option('--password', envvar='DB_PASSWORD', type=str)
@click.option('--database', envvar='DB_DATABASE', type=str)
@click.option('--config-file', default=None)
@click.option('--timeout', default=None, type=float)
@click.option('--interval', default=1, type=int)
@click.option('--timezone', envvar='TZ', default='UTC')
@click.option('--template', type=str, default=None, help='json field template, or @FILENAME in cwd or ~/.sqltail')
@click.option('--get-template', is_flag=True, help='output json field template')
@click.option('--get-columns', is_flag=True)
@click.option('--suffix', type=str, default='_log', help="append SUFFIX to db name (defaults to _log)")
@click.option('-r/-R', '--retry/--no-retry', is_flag=True, default=True, help="retry on database connection failures")
@click.option('-t', '--table', default='log', type=str, help='table name')
@click.option('-c', '--columns', default=None, type=str, help='comma delimited list of output column names')
@click.option('-f', '--filters', default=None, type=str, help='list of filter conditions') 
@click.option('-o', '--output-format', default='json', type=str, help='output format') 
@click.option('-l', '--log-level', envvar="LOG_LEVEL", default='WARNING', type=click.Choice(['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'], case_sensitive=False))

def sqltail(host, port, user, password, database, config_file, timeout, interval, timezone, columns, filters, log_level, get_template, get_columns, table, template, output_format, suffix, retry):

    logging.basicConfig(level=log_level.upper())

    state = None
    while True:
        try:
            db=Database(host, port, user, password, database, config_file=config_file, debug=log_level=='DEBUG', suffix=suffix)
        except DatabaseNotFound as exc:
            if state != type(exc):
                click.echo(str(exc)+' retrying...')
                state = type(exc)
        except DatabaseConnectionFailed as exc:
            if state != type(exc):
                click.echo(str(exc)+' retrying...')
                state = type(exc)
        else:
            state = None
            break
        time.sleep(3)
        if not retry:
            click.exit(1)
        
    columns = columns.split(',') if columns else []
    filters = filters.split(',') if filters else []
    if template:
        if template.startswith('@'):
            template_path = Path(template[1:])
            if not template_path.is_file():
                template_path = Path.home() / '.sqltail' / template[1:]
            template = template_path.read_text()

        columns = from_json(template)

    sql_tail = SQLTail(
        db, 
        table=table,
        fields=columns, 
        filters=filters,
        callbacks=[output],
        tz=timezone,
        interval=interval
    )
    if get_template:
        output(sql_tail.get_field_template(), fmt=output_format)
    elif get_columns:
        output(sql_tail.get_columns(), fmt=output_format)
    else:
        sql_tail.run(timeout=timeout)

def to_json(data):
    if hasattr(data, '__iter__'):
        ret = '[\n  ' + ',\n  '.join([json.dumps(item) for item in data]) + '\n]'
    else:
        ret = json.dumps(data)
    return ret

def from_json(data):
    return json.loads(data) 

def output(msg, fmt=None):
    if fmt=='json':
        msg = to_json(msg)
    click.echo(msg)



# sqltail

from sqltail.monitor import SQLTail
from sqltail.db import Database, DatabaseException, DatabaseConnectionFailed, DatabaseNotFound

__version__='1.0.3'
__license__='Copyright 2021 Reliance Systems, Inc. - MIT license - https://github.com/rstms/sqltail'



import arrow
import datetime
import logging
import time
import copy

"""
ideas:

    Instead of reconnecting and querying for max row ID,
    create a UDF on the server that does the query in a loop and sleeps, only returning data
    when the MAX ID has changed, or a timeout occurs,
    then we don't constantly create/destroy connections and cursors
    we sit in a select, and when the data returns, it's either a timeout or a new max-row-id

"""
    

TZ='UTC'
DATETIME_FIELD_NAMES = ['timestamp', 'created', 'updated']

WAIT_INTERVAL_INIT=0.1
WAIT_INTERVAL_MULTIPLIER=2
WAIT_INTERVAL_MAX=1

class SQLTail():
    def __init__(self, db, table='log', fields=[], filters=[], delimiter=' ', interval=1, callbacks=[print], tz=TZ):

        self.logger=logging.getLogger(__class__.__name__)

        self.db = db
        self.table = table
        self.tz = tz
        self.delimiter = delimiter
        self.interval = interval
        self.callbacks=callbacks
        self.filters = filters
        self.columns = self.get_columns()
        self.fields = self.init_fields(fields)
        self.sql_fields = ','.join([f for f in self.fields])
        self.logger.debug(f"{self}")

    def __str__(self):
        return f"{self.__class__.__name__}<{self.db} {self.table} {self.fields} {self.filters}"

    def init_fields(self, fields):
        ret = dict()
        column_map = {c.Field:c for c in self.columns}
        if not fields:
            fields = list(column_map.keys())
        elif isinstance(fields, str):
            fields = fields.split(',')
        elif isinstance(fields, list):
            if isinstance(fields[0],str):
                pass   # fields may be a list of column names
            elif isinstance(fields[0], dict):
                # fields is a list of field spec dicts
                for spec in fields:
                    field = Field(**spec)
                    ret[field.name]=field
                fields = None
            else:
                TypeError('Unable to interpret fields specification')
        else:
            raise TypeError('fields may be a list of either column names or field specifiers') 

        if fields:
            for field in fields:
                column = column_map[field]
                hint = 'datetime' if column.Type.startswith('datetime') else None
                ret[column.Field] = Field(column.Field, tz=self.tz, type_hint=hint)
        return ret 
            
    def get_columns(self):
        with self.db.cursor() as cursor:
            return cursor.query(f"DESCRIBE {self.table};")

    def get_field_template(self):
        return [field.template() for field in self.fields.values()]

    def run(self, timeout=None):
        self.logger.debug('run: begin')
        if timeout:
            timeout = arrow.utcnow() + datetime.timedelta(seconds=timeout)
        self.running = True
        last_id = self.get_last_row_id()
        wait_interval = WAIT_INTERVAL_INIT
        while self.running:
            self.db.cxn.reconnect()
            self.db.cxn.database = self.db.database
            self.logger.debug(f"Querying new rows since last_id {last_id}...")
            with self.db.cursor() as cursor:
                rows = self.get_new_rows(cursor, last_id)
            if len(rows):
                wait_interval = WAIT_INTERVAL_INIT
                self.logger.debug(f"{len(rows)} row{'' if len(rows)==1 else 's'} returned")
                self.output_rows(rows)
                last_id = rows[-1]._id
            elif timeout and (arrow.utcnow() > timeout):
                self.logger.debug('Timeout')
                self.running = False
            else:
                if wait_interval < WAIT_INTERVAL_MAX:
                    wait_interval *= WAIT_INTERVAL_MULTIPLIER
                else:
                    wait_interval = WAIT_INTERVAL_MAX
                self.logger.debug('No rows returned; sleeping...')
                time.sleep(wait_interval)
        self.logger.debug('run: end')

    def sql_where(self, where=None):
        ret = copy.copy(self.filters)
        if where:
            ret.append(where)
        return f"{'WHERE ' if ret else ''}{' AND '.join(ret)}"

    def get_new_rows(self, cursor, last_id):
        where = self.sql_where(f"id > {last_id}")
        rows = cursor.query(f"SELECT id as _id,{self.sql_fields} FROM {self.table} {where} ORDER BY id;")
        return rows

    def get_last_row_id(self):
        with self.db.cursor() as cursor:
            rows = cursor.query(f"SELECT id from {self.table} {self.sql_where()} ORDER BY id DESC LIMIT 1;")
        return rows[0].id

    def output_rows(self, rows):
        for msg in map(self.format_row, rows):
            for callback in self.callbacks:
                callback(msg)

    def format_row(self, row):
        return self.delimiter.join([self.fields[k].fmt(v) for k,v in row.items() if k in self.fields])
        line = []
        for k,v in row.items():
            if k in self.fields:
                field = self.fields[k]
                func = field.fmt
                text = func(v)
                line.append(text)
        msg = self.delimiter.join(line)
        return msg


class Field():
    def __init__(self, name=None, format_func=None, where_clause=None, truncate=0, left_pad=0, right_pad=0, tz=TZ, type_hint=None):
        self.logger=logging.getLogger(__class__.__name__)
        if not name:
            raise ValueError("Field name cannot be None")
        self.name = name
        self.fmt = self.init_fmt_func(format_func, name, type_hint)
        self.where_clause = where_clause
        self.truncate = truncate
        self.lpad = left_pad
        self.rpad = right_pad
        self.tz = tz 
        self.logger.debug(f"{self}")

    def __str__(self):
        return f"{self.__class__.__name__}<{self.name} {self.fmt.__name__} {self.where_clause} {self.truncate} {self.lpad} {self.rpad} {self.tz}>"

    def template(self):
        return dict( 
                name=self.name, 
                format_func=self.fmt.__name__, 
                where_clause=self.where_clause, 
                truncate=self.truncate,
                left_pad=self.lpad,
                right_pad=self.rpad,
                tz=self.tz
            )

    def init_fmt_func(self, fmt_func, name, type_hint):
        if fmt_func:
            if isinstance(fmt_func, str):
                fmt = getattr(self, fmt_func)
            else:
                fmt = fmt_func
        else:
            if type_hint=='datetime' or name in DATETIME_FIELD_NAMES:
                fmt = self.fmt_datetime
            else:
                fmt = self.fmt_str
        return fmt

    def fmt_datetime(self, dt):
        return arrow.get(dt).to(self.tz).isoformat(' ')[:24]

    def fmt_str(self, value):
        value = str(value)
        if self.lpad:
            value = ' '*(self.lpad-len(value))+value
            if self.truncate:
                value = value[-self.truncate:]
        elif self.rpad:
            value = value+' '*(self.rpad-len(value))
            if self.truncate:
                value = value[:self.truncate]
        elif self.truncate:
            value = value[:self.truncate]
        return value
